##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""
Zope management support for DOM classes.
"""
from AccessControl import ClassSecurityInfo

import Globals
import Acquisition
from Acquisition import aq_parent, aq_base
import App.Management
from OFS.Traversable import Traversable
from DateTime import DateTime # for manage_edit cookie expire
import marshal # for ftp
from Globals import DTMLFile
import DOMProxy
import DOM
import ExtraDOM
import string
from xml.parsers import expat
import xml.dom
import types
from StringIO import StringIO
from NodePath import registry

try:  			# Zope 2.8 style transactions
    import transaction
except ImportError:  	# Old-style transactions
    # BBB: for Zope 2.7
    class TransactionMan:
        def begin(self):                get_transaction().begin()
        def commit(self, sub=False):    get_transaction().commit(sub)
        def abort(self, sub=False):     get_transaction().abort(sub)
        def get(self):                  return get_transaction()
    transaction = TransactionMan()

parserr=('Sorry, an XML parsing error occurred.  Check '
         'your XML document for well-formedness and try '
         'to upload it again after modification.<br><br>')

# 
# Management mixin classes
#
   
class DOMTraversable(Traversable):
    "Mixin class for DOM classes to provide Zope traversability."

    security = ClassSecurityInfo()

    security.declarePublic('getPhysicalPath')
    def getPhysicalPath(self):
        """Returns a path that can be used to access this object again
        later."""
        # first get path to persistent document
        doc = self._persistentDoc._earlyAqChain()
        path = doc.getPhysicalPath()
        # then get path to node and attach it to that path steps
        nodepath = registry.create_path(doc, self, 'child')
        if nodepath:
            return path + (nodepath,)
        else:
            return path
        
    security.declarePublic('getNodePath')
    def getNodePath(self, scheme_name):
        """Create a node path for this node.
        FIXME: this has the same name but other signature as the
        one in ParsedXML for document.
        """
        return registry.create_path(self._persistentDoc._earlyAqChain(),
                                    self, scheme_name)
    
    # sequence interface, for backwards compatibility.
    # normally the document's __getitem__ takes care of this
    def __getitem__(self, key):
        return self.childNodes[int(key)].__of__(self)

    # other helpers
    def _earlyAqChain(self):
        """Return the earliest ref to self in self's aq_chain.  Helper function
        for getPhysicalPath."""
        chain = self.aq_chain
        chain.reverse()
        for parent in chain:
            if parent == self:
                return parent
        return self

Globals.InitializeClass(DOMTraversable)

class DOMPublishable(DOMTraversable):
    "Mixin class for DOM classes to provide Zope publishability."
    # tree tag methods
    
    def tpURL(self):
        """
        Return a string used for an URL relative to parent.  Used by
        the dtml-tree tag.
        """
        i = 0
        node = self.getDOMObj().previousSibling
        while node:
            i = i + 1
            node = node.previousSibling
        return str(i)

    def tpValues(self):
        "Return a list of immediate subobjects.  Used by the dtml-tree tag."
        retList = [] # dtml-tree wants to own this
        for i in self.childNodes:
            retList.append(i)
        return retList

    def tpId(self):
        "Return a value to be used as an id in tree state."
        return self.tpURL()

    # partial ObjectManagerItem interface

    def getId(self):
        "Return the id of the object as a string."
        base = aq_base(self)
        name=getattr(base, 'id', None)
        if name is not None:
            return name
        return self.tpURL()

    # I don't want to add PrincipiaSearchSource, because it'd be expensive
    # to trigger printing every node in a document.  Look into this when
    # we have a better caching plan.        

    # Partial ObjectManager interface.  Nodes are *not* ObjectManagers,
    # these are for other Zope tools to be useful.

    def objectValues(self, spec=None):
        """
        Returns a list of actual subobjects of the current object.
        If 'spec' is specified, returns only objects whose meta_type
        match 'spec'.
        """
        if spec is not None:
            if isinstance(spec, type('s')):
                spec=[spec]
            set=[]
            for ob in self.childNodes:
                if ob['meta_type'] in spec:
                    set.append(ob)
            return set
        return list(self.childNodes)
    
    def objectIds(self, spec=None):
        """
        Returns a list of subobject ids of the current object.
        If 'spec' is specified, returns objects whose meta_type
        matches 'spec'.
        """
        return map(lambda i: i.getId(), self.objectValues(spec))

    def objectItems(self, spec=None):
        """
        Returns a list of (id, subobject) tuples of the current object.
        If 'spec' is specified, returns only objects whose meta_type match
        'spec'
        """
        r=[]
        a=r.append
        for ob in self.objectValues(spec): a((ob.getId(), ob))
        return r

    # FTP interface
    
    def manage_FTPget(self):
        """Returns the source content of an object. For example, the
        source text of a Document, or the data of a file."""
        return self.__str__()        

class DOMIO:
    "Mixin class for DOM classes to provide parsing, writing."
    
    def writeStream(self, stream=None, encoding=None, prettyPrint=0):
        "Write the XML representation of this object to stream."
        # work thru DOM object to avoid making proxy nodes
        return ExtraDOM.writeStream(self.getDOMObj(), stream, encoding,
                                    prettyPrint=prettyPrint)
    
    def __str__(self):
        "Return the XML representation of this object."
        return self.writeStream().getvalue()

    # This is a workaround for newer browser behavior, and will probably
    # break older browsers.  Recent browsers put '<' in the textarea when the
    # source is '&lt;", and when the form is sent they send '<', and so on.
    # This quotes the relevent refs; when the textarea is sent back the
    # browser unquotes them, so the end result is WYSIWYG.  Yuck!
    # If a browser doesn't do this __str__ should be used.
    def textareaStr(self, prettyPrint=0):
        """Return the XML representation of this object in a format safe for
        a textarea, with certain entity references quoted."""
        # FIXME
        # perhaps we should output text in the encoding the document
        # was saved in, but this is too complicated for now, so we just
        # output UTF-8. Unfortunately this is not very pretty in the browser
        # unless you set it to display in UTF-8.
        outStr = self.writeStream(
            encoding=None, prettyPrint=prettyPrint).getvalue().encode('UTF-8')
        outStr = string.replace(outStr, '&amp;', '&amp;amp;')        
        outStr = string.replace(outStr, '&lt;', '&amp;lt;')
        outStr = string.replace(outStr, '&gt;', '&amp;gt;')
        outStr = string.replace(outStr, '<', '&lt;')
        outStr = string.replace(outStr, '>', '&gt;') 
        return outStr

    def index_html(self, REQUEST=None, RESPONSE=None):
        "Returns publishable source according to content type"
        # set content type header for raw XML if necessary
        if RESPONSE:
            if self._persistentDoc:
                contentType = self._persistentDoc.contentType
            else:
                contentType = 'text/xml'
            contentType = contentType + ';charset=utf-8'
            RESPONSE.setHeader('Content-Type', contentType)
        return self.writeStream(None).getvalue()

    # Parsing a node removes self & subtree from document; users of other
    # refs to subnodes have to take care to notice this.
    def parseXML(self, file):
        """Parse file as XML, replace myself with the resulting tree,
        return node replacing self."""
        doc = self._persistentDoc
        namespaces = not (doc and doc.noNamespaces) # default true
        node = ExtraDOM.parseFile(self.getDOMObj(), file, namespaces)
        if self.nodeType == xml.dom.Node.DOCUMENT_NODE:
            errorStr = ("Parsing at the document node must be done on the "
                        "persistent proxy node, which can't be found for some "
                        "reason.\n"
                        "Traverse to the persistent node and try there.")
            raise RuntimeError, errorStr
        else:
            return doc.wrapDOMObj(node) # proxy and node not part of tree


class DOMManageable(DOMIO, DOMPublishable, App.Management.Tabs):
    "Mixin class for DOM classes to provide Zope management interfaces."

    security = ClassSecurityInfo()
    security.setPermissionDefault('View DOM hierarchy', ['Manager'])    
    
    # the UI of ParsedXML is UTF-8
    management_page_charset = 'UTF-8'

    security.declareProtected('View management screens',
                              'manage_editForm')
    manage_editForm = Globals.DTMLFile('dtml/transEdit', globals(),
                                       __name__='manage_editForm')
    security.declareProtected('View DOM hierarchy',
                              'manage_DOMTree')
    manage_DOMTree = Globals.DTMLFile('dtml/DOMTree', globals(),
                                      __name__='manage_DOMTree')
    
    manage_options = (
        {'label':'Edit', 'action':'manage_editForm',
         'help': ('ParsedXML', 'ParsedXML_Edit.stx')},
        {'label':'DOM', 'action':'manage_DOMTree',
         'help': ('ParsedXML', 'ParsedXML_DOM.stx')},
        {'label':'Raw', 'action':'index_html'})
    
    def makeErrorOutput(self, data, offset, lineno):
        """return a HTML-renderable output of data with a text pointer
        to the position at offset, lineno"""
        # make pointer
        pointerLine = ('<font color="red">' +
                       '-' * (offset - 1) + '^' + '\n\n' +
                       '</font>')
        # add rest of data
        # the parser saw a normalized version; normalize data
        data = string.replace(data, "\r\n", "\n")
        data = string.replace(data, "\r", "\n")
        # make reconstruced data rep with break at error
        dataA = ''
        dataB = ''
        # add earlier data
        for line in range(lineno - 1):
            dataA = dataA + data[0:string.index(data, '\n') + 1]
            data = data[string.index(data, '\n') + 1:]
        dataA = dataA + data[0:offset] + '\n\n'
        # add later data
        dataB = dataB + ' ' * (offset)
        dataB = dataB + data[offset:] + '\n'
        # strip brackets, format
        dataA = string.replace(dataA, "<", "&lt;")
        dataA = string.replace(dataA, ">", "&gt;")
        dataB = string.replace(dataB, "<", "&lt;")
        dataB = string.replace(dataB, ">", "&gt;")
        dataOut = ('<div align=left><br><br><pre>' +
                   dataA + pointerLine + dataB +
                   '</pre><br><br></div>')
        return dataOut
    
    # dict to help notice, handle size change request in manage_editForm
    _size_changes={
        'Bigger': (5,5),
        'Smaller': (-5,-5),
        'Narrower': (0,-5),
        'Wider': (0,5),
        'Taller': (5,0),
        'Shorter': (-5,0),
        }

    # helper function to set size cookies and return edit form
    def _er(self, data, title, contentType,
            SUBMIT, dtpref_cols, dtpref_rows, REQUEST):
        dr,dc = self._size_changes[SUBMIT]
        rows=max(1,string.atoi(dtpref_rows)+dr)
        cols=max(40,string.atoi(dtpref_cols)+dc)
        e=(DateTime('GMT') + 365).rfc822()
        resp=REQUEST['RESPONSE']
        resp.setCookie('dtpref_rows',str(rows),path='/',expires=e)
        resp.setCookie('dtpref_cols',str(cols),path='/',expires=e)
        return self.manage_editForm(self, REQUEST,
                                dtpref_cols = cols, dtpref_rows = rows)

    security.declareProtected('Edit ParsedXML', 'manage_edit')
    def manage_edit(self, data, title = '', contentType = None,
                    useNamespaces = 1,
                    SUBMIT = 'Change',
                    dtpref_cols = '50', dtpref_rows = '20', REQUEST = None):
        """
        If SUBMIT is a size pref variable, handle a textarea size change.
        Otherwise parse the given text and handle the result.
        """
        # just get back to the dtml if we're changing size
        if self._size_changes.has_key(SUBMIT):
            return self._er(data, title, contentType,
                            SUBMIT, dtpref_cols, dtpref_rows, REQUEST)
        # if we are pretty printing, just redraw the form
        if SUBMIT == 'PrettyPrint':
            return self.manage_editForm(
                self, REQUEST,
                textareaStr2=self.textareaStr(prettyPrint=1))
                                        
        if (hasattr(self, 'nodeType')
            and self.nodeType == xml.dom.Node.DOCUMENT_NODE):
            #if we're the main doc
            self.title = str(title)
            if contentType:
                self.contentType = str(contentType)
            self.noNamespaces = not useNamespaces        
        
        f = StringIO(data)
        try:
            newNode = self.parseXML(f) # self not in doc if this succeeds
        except expat.error, e:
            transaction.abort()
            if REQUEST:
                dataOut = self.makeErrorOutput(data, e.offset, e.lineno)
                err = "%s%s%s" % (parserr, '<font color="red">%s</font>'
                                % getattr(e, 'args', ''), dataOut)
                return Globals.MessageDialog(
                    title = 'XML Parsing Error',
                    message = err,
                    action = 'manage_editForm')
            raise
        if REQUEST:
            message = "Saved changes."
            # wish I knew why we have to set textareaStr for the form
            return newNode.manage_editForm(self, REQUEST,
                                           textareaStr=newNode.textareaStr(),
                                           management_view="Edit",
                                           manage_tabs_message=message)

    security.declareProtected('Edit ParsedXML', 'manage_upload')
    def manage_upload(self, file, REQUEST=None):
        "Parse the given file and handle the result."
        try:
            newNode = self.parseXML(file)
        except expat.error, e:
            transaction.abort()
            if REQUEST:
                file.seek(0)
                dataOut = self.makeErrorOutput(file.read(),
                                               e.offset, e.lineno)
                err = "%s%s%s" % (parserr, '<font color="red">%s</font>'
                                % getattr(e, 'args', ''), dataOut)
                return Globals.MessageDialog(
                    title  = 'XML Parsing Error',
                    message = err,
                    action = 'manage_editForm')
            raise
        if REQUEST:
            return newNode.manage_editForm(
                self, REQUEST,
                manage_tabs_message='Saved changes.')

Globals.InitializeClass(DOMManageable)

#
# And finally, classes to mix management and DOM proxies.
#

class ManageableWrapper:
    """
    Mixin class to go alongside ManageableNode classes.
    Provides the wrapDOMObj function to create ManageableNode classes.
    """

    # anything that returns subobjs must grant access to them
    __allow_access_to_unprotected_subobjects__ = 1

    def wrapNamedNodeMap(self, obj):
        if obj is None:
            return None
        parent = aq_parent(self) or self
        return ManageableNamedNodeMap(obj, self._persistentDoc).__of__(parent)

    def wrapNodeList(self, obj):
        parent = aq_parent(self) or self
        return ManageableNodeList(obj, self._persistentDoc).__of__(parent)
    
    def wrapDOMObj(self, node):
        """Return the appropriate manageable class wrapped around the node."""
        if node is None:
            return
        wrapper_type = WRAPPER_TYPES[node._get_nodeType()]
        parent = aq_parent(self) or self
        return wrapper_type(node, self._persistentDoc).__of__(parent)
        
# According to DOM Erratum Core-14, the empty string should be
# accepted as equivalent to null for hasFeature().

_MANAGEABLE_DOM_FEATURES = (
    ("org.zope.dom.persistence", None),
    ("org.zope.dom.persistence", ""),
    ("org.zope.dom.persistence", "1.0"),
    ("org.zope.dom.acquisition", None),
    ("org.zope.dom.acquisition", ""),
    ("org.zope.dom.acquisition", "1.0"),
    )

_MANAGEABLE_DOM_NON_FEATURES = (
    ("load", None),
    ("load", ""),
    ("load", "3.0"),
    )

class ManageableDOMImplementation(DOMProxy.DOMImplementationProxy):
    """A proxy of a DOMImplementation node that defines createDocument
    to return a ManageableDocument.
    """
    
    def hasFeature(self, feature, version):
        feature = string.lower(feature)
        if (feature, version) in _MANAGEABLE_DOM_FEATURES:
            return 1
        if (feature, version) in _MANAGEABLE_DOM_NON_FEATURES:
            return 0
        return DOMProxy.DOMImplementationProxy.hasFeature(self, feature,
                                                          version)

    def createDocumentType(self, qualifiedName, publicId, systemId):
        DOMDocumentType = self._createDOMDocumentType(qualifiedName,
                                                      publicId, systemId)
        return ManageableDocumentType(DOMDocumentType)
    
    def createDocument(self, namespaceURI, qualifiedName, docType=None):
        if docType is not None:
            if docType.ownerDocument is not None:
                raise xml.dom.WrongDocumentErr
            mdocType = docType.getDOMObj()
        else:
            mdocType = None
        DOMDocument = self._createDOMDocument(namespaceURI, qualifiedName,
                                              mdocType)
        return ManageableDocument(DOMDocument, DOMDocument)

theDOMImplementation = ManageableDOMImplementation()

# XXX We're implicitly acquiring so we can get ZopeTime (and probably a
# jillion other things) in our DTML methods.  We should be explicit.
class ManageableNode(ManageableWrapper, DOMProxy.NodeProxy, DOMManageable,
                     Acquisition.Implicit):
    "A wrapper around a DOM Node."

    # this is mainly here to make later inheritance safer
    def __init__(self, node, persistentDocument):
        # inherit from DOMProxy.NodeProxy
        ManageableNode.inheritedAttribute('__init__')(self, node,
                                                      persistentDocument)

class ManageableNodeList(ManageableWrapper, DOMProxy.NodeListProxy,
                         Acquisition.Implicit):
    "A wrapper around a DOM NodeList."
    meta_type = "Manageable NodeList"

    # redefine to get back the [] syntax with acquisition, eh?
    def __getslice__(self, i, j):
        return self.wrapNodeList(self._node.__getslice__(i,j))

    # redefine to get back the [] syntax with acquisition, eh?
    def __getitem__(self, i):
        return self.wrapDOMObj(self._node.__getitem__(i))        

class ManageableNamedNodeMap(ManageableWrapper, DOMProxy.NamedNodeMapProxy,
                             Acquisition.Implicit):
    "A wrapper around a DOM NamedNodeMap."
    meta_type = "Manageable NamedNodeMap"

    # redefine to get back the [] syntax with acquisition, eh?
    def __getitem__(self, i):
        return self.wrapDOMObj(self._node.__getitem__(i))        
    
class ManageableDocumentFragment(ManageableWrapper,
                                 DOMProxy.DocumentFragmentProxy,
                                 ManageableNode):
    "A wrapper around a DOM DocumentFragment."
    meta_type = "Manageable Document Fragment"
    
class ManageableElement(ManageableWrapper, DOMProxy.ElementProxy,
                        ManageableNode):
    "A wrapper around a DOM Element."
    meta_type = "Manageable Element"

class ManageableCharacterData(ManageableWrapper,
                              DOMProxy.CharacterDataProxy, ManageableNode):
    "A wrapper around a DOM CharacterData."
    meta_type = "Manageable Character Data"

class ManageableCDATASection(ManageableWrapper,
                             DOMProxy.CDATASectionProxy, ManageableNode):
    "A wrapper around a DOM CDATASection."
    meta_type = "Manageable CDATASection"

class ManageableText(ManageableWrapper,
                     DOMProxy.TextProxy, ManageableCharacterData):
    "A wrapper around a DOM Text."
    meta_type = "Manageable Text"

class ManageableComment(ManageableWrapper,
                        DOMProxy.CommentProxy, ManageableCharacterData):
    "A wrapper around a DOM Comment."
    meta_type = "Manageable Comment"    

class ManageableProcessingInstruction(ManageableWrapper,
                                      DOMProxy.ProcessingInstructionProxy,
                                      ManageableNode):
    "A wrapper around a DOM ProcessingInstruction."
    meta_type = "Manageable Processing Instruction"    

class ManageableAttr(ManageableWrapper, DOMProxy.AttrProxy, ManageableNode):
    "A wrapper around a DOM Attr."
    meta_type = "Manageable Attr"    

#ManageableDocument is not necessarily a persistent object, even when a
#persistent subclass such as ParsedXML has been instantiated.  Traversing
#up to the document can create a new transient proxy.  Persistent attributes
#must be set on the persistent version.
class ManageableDocument(ManageableWrapper, DOMProxy.DocumentProxy,
                         ManageableNode):
    "A wrapper around a DOM Document."
    meta_type = "Manageable Document"    

    implementation = theDOMImplementation

    def __init__(self, node, persistentDocument):
        ManageableNode.__init__(self, node, persistentDocument)

    def _get_implementation(self):
        return self.implementation
    
    #block set of implementation, since we don't proxy it the same
    def __setattr__(self, name, value):
        if name == "implementation":
            raise xml.dom.NoModificationAllowedErr()
        ManageableDocument.inheritedAttribute('__setattr__')(self, name, value)

# DOM extended interfaces

class ManageableEntityReference(ManageableWrapper,
                                DOMProxy.EntityReferenceProxy,
                                ManageableNode):
    "A wrapper around a DOM EntityReference."
    meta_type = "Manageable Entity Reference"

class ManageableEntity(ManageableWrapper, DOMProxy.EntityProxy,
                       ManageableNode):
    "A wrapper around a DOM Entity."
    meta_type = "Manageable Entity"

class ManageableNotation(ManageableWrapper, DOMProxy.NotationProxy,
                         ManageableNode):
    "A wrapper around a DOM Notation."
    meta_type = "Manageable Notation"

class ManageableDocumentType(ManageableWrapper, DOMProxy.DocumentTypeProxy,
                             ManageableNode):
    "A wrapper around a DOM DocumentType."
    meta_type = "Manageable Document Type"    


Node = xml.dom.Node
WRAPPER_TYPES = {
    Node.ELEMENT_NODE: ManageableElement,
    Node.ATTRIBUTE_NODE: ManageableAttr,
    Node.TEXT_NODE: ManageableText,
    Node.CDATA_SECTION_NODE: ManageableCDATASection,
    Node.ENTITY_REFERENCE_NODE: ManageableEntityReference,
    Node.ENTITY_NODE: ManageableEntity,
    Node.PROCESSING_INSTRUCTION_NODE: ManageableProcessingInstruction,
    Node.COMMENT_NODE: ManageableComment,
    Node.DOCUMENT_NODE: ManageableDocument,
    Node.DOCUMENT_TYPE_NODE: ManageableDocumentType,
    Node.DOCUMENT_FRAGMENT_NODE: ManageableDocumentFragment,
    Node.NOTATION_NODE: ManageableNotation,
    }
del Node



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"tests to make sure that DOM objects support truth testing"

import os, sys
  
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Testing import ZopeTestCase

ZopeTestCase.installProduct('ParsedXML')

from Products.ParsedXML import ParsedXML, DOM

from operator import truth

class WrappedTruthableTestCaseBase(ZopeTestCase.ZopeTestCase):

    def afterSetUp(self):
        self.document = doc = ParsedXML.ParsedXML('foo')
        self.floating_element = doc.createElement("per")
        self.attached_element = doc.documentElement

class DOMTruthableTestCaseBase(ZopeTestCase.ZopeTestCase):

    def afterSetUp(self):
        self.document = doc = DOM.theDOMImplementation.createDocument(
            None, 'root', None)
        self.floating_element = doc.createElement("per")
        self.attached_element = doc.documentElement

class TruthableTestCaseTests:

    def testFloatingTruthable(self):
        assert truth(self.floating_element) == 1

    def testDOMTruthable(self):
        assert truth(self.document) == 1
        assert truth(self.document.documentElement.parentNode) == 1

    def testAttachedTruthable(self):
        assert truth(self.attached_element) == 1

class DOMTruthableTestCase(DOMTruthableTestCaseBase,
                           TruthableTestCaseTests):
    pass

class WrappedTruthableTestCase(WrappedTruthableTestCaseBase,
                               TruthableTestCaseTests):
    pass

if __name__ == '__main__':
    framework()
else:
    import unittest
    def test_suite():
        suite = unittest.TestSuite()
        suite.addTest(unittest.makeSuite(DOMTruthableTestCase))
        suite.addTest(unittest.makeSuite(WrappedTruthableTestCase))
        return suite



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""
Zope Product creation for Parsed XML.
"""

from OFS.SimpleItem import SimpleItem
from AccessControl import ClassSecurityInfo
from AccessControl.Role import RoleManager
from Persistence import Persistent
from Acquisition import Implicit
from OFS.Cache import Cacheable
import Globals
from Globals import DTMLFile
from ManageableDOM import ManageableDocument, DOMManageable, \
     theDOMImplementation, parserr
from StringIO import StringIO
from xml.parsers import expat
import DOM, ExtraDOM
import helpers
from types import StringType
from NodePath import registry

MARKER = []  # dummy default object for cache return

contentTypes = ['text/html', 'application/html', 'text/xml']
    
class ParsedXML(SimpleItem, ManageableDocument, Cacheable):
    "The Parsed XML top object, and the persistent head of the tree."
     
    meta_type = 'Parsed XML'

    security = ClassSecurityInfo()
    
    manage_options = (DOMManageable.manage_options +
                      RoleManager.manage_options +
                      Cacheable.manage_options)

    security.setPermissionDefault('Edit ParsedXML', ['Manager'])

    security.declareProtected('Edit ParsedXML',
                              'manage_editForm')
    manage_editForm = DTMLFile('dtml/persEdit', globals(),
                               __name__='manage_editForm')

    icon = 'misc_/ParsedXML/pxml.gif'

    def __init__(self, id, XMLstring = None, namespaces = 1,
                 contentType = "text/xml"):
        "Initialize a Parsed XML object"
        self.id = id
        self._persistentDoc = self # used by transient proxies
        self.noNamespaces = not namespaces
        if contentType not in contentTypes:
            import string
            raise RuntimeError, (
                "Bad content type %s; valid types are %s"
                % (str(contentType), string.join(contentTypes)))
        self.contentType = contentType
        self.initFromDOMDocument(createDOMDocument(XMLstring, namespaces))
        if XMLstring:
            self._lenCache = len(XMLstring)
        else:
            self._lenCache = len(str(self))
        self._lenCorrect = 1

    security.declarePrivate('initFromDOMDocument')
    def initFromDOMDocument(self, DOMdoc):
        "Initialize a Parsed XML from a DOM Document"
        # inherit from ManageableDocument
        ParsedXML.inheritedAttribute('__init__')(self, DOMdoc, self)

    def manage_afterAdd(self, object, container):
        """Store the container that we're being added to.  This is used
        to traverse back to the persistent document."""
        self._container = container

    #
    # methods that deal with persistence
    #

    # DOM nodes use __changed__ because its acquirable
    # Setting _p_changed is equivalent.
    def __changed__(self, *args):
        """Override the acquired __changed__ method used by the non-DB
        DOM nodes, update length cache, and mark this object as dirty."""
        if getattr(self, '_v_lenCorrect', None):
            self._lenCorrect = 1
        else:
            self._lenCorrect = 0
        self.ZCacheable_invalidate()            
        self._p_changed = 1

    # wrap DOMIO to provide cacheing
    security.declareProtected('View management screens', 'index_html')
    def index_html(self, REQUEST = None, RESPONSE = None):
        "Returns publishable source according to content type"
        if RESPONSE:
            RESPONSE.setHeader('Content-type', self.contentType)
        data = self.ZCacheable_get(default = MARKER)
        if data is not MARKER:
            return data
        # inherit from DOMIO
        data = ParsedXML.inheritedAttribute('index_html')(self,
                                                          REQUEST, RESPONSE)
        self.ZCacheable_set(data)
        return data

    security.declareProtected('View management screens', 'get_size')
    def get_size(self):
        "Length of the XML string representing this node in characters."
        if not getattr(self, '_lenCorrect', 0):
            self._lenCache = len(str(self))
            self._lenCorrect = 1
        return self._lenCache

    # methods that override ZDOM methods that are incorrect
    #

    security.declareProtected('Access contents information',
                              'getElementsByTagName')
    def getElementsByTagName(self, tagName):
        return self.__getattr__("getElementsByTagName")(tagName)

    security.declareProtected('Access contents information',
                              'hasChildNodes')
    def hasChildNodes(self):
        return self.__getattr__("hasChildNodes")()

    # 
    # methods that override SimpleItem; sigh, multiple inheritance.
    #
    security.declareProtected('Access contents information', 'objectValues')
    def objectValues(self, spec=None):
        """
        Returns a list of actual subobjects of the current object.
        If 'spec' is specified, returns only objects whose meta_type
        match 'spec'.
        """
        return ManageableDocument.objectValues(self, spec)

    security.declareProtected('Access contents information', 'objectIds')
    def objectIds(self, spec=None):
        """
        Returns a list of subobject ids of the current object.
        If 'spec' is specified, returns objects whose meta_type
        matches 'spec'.
        """
        return ManageableDocument.objectIds(self, spec)
    
    security.declareProtected('Access contents information', 'objectItems')
    def objectItems(self, spec=None):
        """
        Returns a list of (id, subobject) tuples of the current object.
        If 'spec' is specified, returns only objects whose meta_type match
        'spec'
        """
        return ManageableDocument.objectItems(self, spec)
    
    security.declareProtected('Access contents information', 'tpValues')   
    def tpValues(self):
        "Return a list of immediate subobjects.  Used by the dtml-tree tag."
        return ManageableDocument.tpValues(self)

    # override ManageableDocument's method; we can't persist new DOM node by
    # hanging off of parents
    security.declareProtected('Edit ParsedXML', 'parseXML')
    def parseXML(self, file):
        "parse file as XML, replace DOM node with resulting tree, return self"
        namespaces = not self.noNamespaces
        node = ExtraDOM.parseFile(self.getDOMObj(), file, namespaces)
        self.initFromDOMDocument(node)
        self.__changed__(1)
        return self

    security.declareProtected('Access contents information',
                              'getDOM')
    def getDOM(self):
        """Get the Document node of the DOM tree.
        """
        return self

    security.declareProtected('Access contents information',
                              'getNodePath')
    def getNodePath(self, scheme_name, node):
        """Create the node path for a particular node in the tree.
        """
        # if we're asking for node of this document itself
        if node is self:
            return 'scheme_name'
        # otherwise ask for nodepath of node
        return node.getNodePath(scheme_name)

    security.declareProtected('Access contents information',
                              'resolveNodePath')
    def resolveNodePath(self, path):
        """Resolve node path from top of the tree to node.
        """
        # start resolving from the document node
        doc = self._persistentDoc._earlyAqChain()
        # FIXME: could use raw DOM instead of management wrappers
        return registry.resolve_path(doc, path)
        
    def __getitem__(self, s):
        """Handle node paths.
        """
        # backwards compatibility -- handle classic 0/1/2 paths
        try:
            return self.childNodes[int(s)].__of__(self)
        except ValueError:
            pass
        
        # start resolving from the document node
        doc = self._persistentDoc._earlyAqChain()
        # FIXME: could use raw DOM instead of management wrappers
        result = registry.resolve_path(doc, s)
        # FIXME: does this convince ZPublisher to show NotFound?
        if result is None:
            raise KeyError, "Could not resolve node path."
        return result

Globals.InitializeClass(ParsedXML)

def createDOMDocument(XMLstring = None, namespaces = 1):
    "Helper function to create a DOM document, without any proxy wrappers."
    if XMLstring:
        XMLstring=StringIO(XMLstring)
        # more efficient to not use ExtraDOM here
        return DOM.ExpatBuilder.parse(XMLstring, namespaces)
    # we use DOM.theDOMImplementation, not ManageableDOMs, for efficiency
    return DOM.theDOMImplementation.createDocument(
        None, "mydocument", None)

def manage_addParsedXML(context, id, title='', file='',
                        useNamespaces=1, contentType="text/xml",
                        REQUEST=None):
    "Add a Parsed XML instance with optional file content."
    
    if not file or not isinstance(file, StringType):
        file ='<?xml version = "1.0"?><emptydocumentElement/>'
    try:
        ob = ParsedXML(id, file, useNamespaces, contentType)
    except expat.error, e:
        if REQUEST is not None:
            err = "%s%s" % (parserr, '<font color="red">%s</font>'
                            % getattr(e, 'args', ''))
            return Globals.MessageDialog(
                title= 'XML Parsing Error',
                message = err,
                action='manage_main')
        raise
    ob.title = str(title)
    context._setObject(id, ob)
    helpers.add_and_edit(context, id, REQUEST, 'manage_editForm')

manage_addParsedXMLForm = DTMLFile('dtml/documentAdd', globals(),
                                   __name__='manage_addParsedXMLForm')



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

# DB stuff copied from testBTrees.py

import sys,os

if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Testing import ZopeTestCase

ZopeTestCase.installProduct('ParsedXML')

from Products.ParsedXML import ParsedXML

import glob

from Products.ParsedXML import ParsedXML
from StringIO import StringIO

class PersistenceTestCase(ZopeTestCase.ZopeTestCase):

    implementation = ParsedXML.theDOMImplementation

    def openDB(self):
        from ZODB.FileStorage import FileStorage
        from ZODB.DB import DB
        storage = FileStorage(self.dbName)
        db = DB(storage)
        self.db = db.open().root()

    def closeDB(self):
        get_transaction().commit()
        self.document = None
        self.db._p_jar._db.close()
        self.db = None

    def getAppDocument(self):
        "put the document that startup put in the db in self.document"
        self.document = self.db['doc']

    def cycleDB(self):
        """close and open the db and replace self.document.
        Any nonpersistent changes to self.document should be lost."""
        self.closeDB()
        self.openDB()
        self.getAppDocument()

    def delDB(self):
        map(os.unlink, glob.glob("fs_tmp__*"))

    def afterSetUp(self):
        """open db, create a document in the db, set self.document to it"""

        self.dbName = 'fs_tmp__%s' % os.getpid()
        self.openDB()
        self.db['doc'] = ParsedXML.ParsedXML('foo')
        get_transaction().commit()
        self.document = self.db['doc']

    def beforeTearDown(self):
        self.closeDB()
        self.delDB()

    def testIDPersistence(self):
        "assert that changing the ID persists over transactions"
        self.document._setId('newId')
        self.cycleDB()
        assert self.document.getId() == 'newId'

    def testParsedXMLDOMPersistence(self):
        "assert that a Parsed XML DOM edit persists over transactions"
        elt = self.document.createElement('elt')
        self.document.firstChild.appendChild(elt)
        self.cycleDB()
        childLen = self.document.firstChild.childNodes.length
        assert childLen == 1, "DOM edit didn't persist"

    def testDOMPersistence(self):
        "assert that a DOM edit persists over transactions"

        doc = ParsedXML.createDOMDocument()

        import OFS.SimpleItem
        si = self.db['si'] = OFS.SimpleItem.SimpleItem()
        si.doc = doc
        elt = si.doc.createElement('elt')
        si.doc.firstChild.appendChild(elt)

        self.cycleDB()
        # not using the normal db document, must grab ourselves
        si = self.db['si']
                
        childLen = si.doc.firstChild.childNodes.length
        assert childLen == 1, "DOM edit didn't persist"
        
    def testParsePersistence(self):
        "assert that a parse persists over transactions"
        testDir = os.path.join(
            sys.modules['Products.ParsedXML'].__path__[0],
            'tests')
        filename = os.path.join(testDir, 'xml', '4ohn4ktj.xml')
        file = open(filename)
        self.document.parseXML(filename)
        file.close()
        childLen = self.document.firstChild.childNodes.length
        self.cycleDB()
        childLen1 = self.document.firstChild.childNodes.length
        assert childLen == childLen1, "parse didn't persist"

    def testSubnodeParsePersistence(self):
        "assert that a subnode parse persists over transactions"        
        docString = '<?xml version="1.0" ?><foo></foo>'
        subNodeString = '<foo>bar</foo>'
        self.document.documentElement.parseXML(StringIO(subNodeString))
        self.cycleDB()
        childLen = self.document.documentElement.childNodes.length
        assert childLen == 1, "parse on subnode didn't persist"

    # right now it's too annoying to get parsing to work at a transient
    # document proxy, but it could be less aggravating if we had a better
    # way to get at the persistent document.
    #def checkTransientDocumentParsePersistence(self):
    #    """assert that a parse of a transient proxy of the document
    #    persists over transactions"""
    #    docStringBefore = '<?xml version="1.0" ?><foo></foo>'
    #    docStringAfter = '<?xml version="1.0" ?><foo>bar</foo>'
    #    self.document.documentElement.ownerDocument.parseXML(
    #        StringIO(docStringAfter))
    #    self.cycleDB()
    #    childLen = self.document.documentElement.childNodes.length
    #    assert childLen == 1, "parse on subnode didn't persist"

#      def checkTheseDamnTests(self):
#          "assert that I understand how these tests should work"
#          docString = '<?xml version="1.0" ?><foo>fff</foo>'        
#          self.document.parseXML(StringIO(docString))
#          get_transaction().commit()
#          # db and doc length now 1
#          self.document.documentElement.removeChild(
#              self.document.documentElement.firstChild)
#          # doc length 0, db length 1
        
#          # do what closeDB and openDB do, but *don't commit*
#          # so the above edit shouldn't stay.

#          #closeDB but don't commit
#          #get_transaction().commit()
#          self.document = None
#          self.db._p_jar._db.close()
#          self.db = None

#          #openDB
#          from ZODB.FileStorage import FileStorage
#          from ZODB.DB import DB
#          storage = FileStorage(self.dbName)
#          db = DB(storage)
#          self.db = db.open().root()

#          self.getAppDocument()
#          get_transaction().commit()
#          assert self.document.documentElement.childNodes.length == 1, (
#              "these tests are faulty")


if __name__ == '__main__':
    framework()
else:
    import unittest
    def test_suite():
        suite = unittest.TestSuite()
        suite.addTest(unittest.makeSuite(PersistenceTestCase))
        return suite



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import string

class NodePathError(Exception):
    pass

class NodePathSchemeRegistry:
    def __init__(self):
        self._schemes = {}
        
    def register_scheme(self, scheme):
        self._schemes[scheme._scheme_name] = scheme

    def resolve_steps(self, top_node, steps, scheme_name):
        """Resolve steps from top_node.
        """
        # if we don't know scheme, return None as node could not be found
        if not self._schemes.has_key(scheme_name):
            raise NodePathError, "Unknown scheme: %s" % scheme_name 
        # we do know the scheme, so look it up
        scheme = self._schemes[scheme_name]
        # now resolve steps using this scheme
        return scheme.resolve_steps(top_node, steps)
    
    def resolve_path(self, top_node, path):
        """Resolve path.
        """
        if not path:
            return top_node
        steps = string.split(path, ',')
        # get scheme as first element of path
        scheme_name, steps = steps[0], steps[1:]
        # resolve steps using scheme
        return self.resolve_steps(top_node, steps, scheme_name)

    def create_steps(self, top_node, node, scheme_name):
        """Construct steps tuple to node.
        """
        return self._schemes[scheme_name].create_steps(top_node, node)
    
    def create_path(self, top_node, node, scheme_name):
        """Construct path to node according to scheme_name.
        """
        steps = self.create_steps(top_node, node, scheme_name)
        if not steps:
            return '' 
        return '%s,%s' % (scheme_name, string.join(steps, ','))

class BaseNodePathScheme:
    def __init__(self, scheme_name):
        self._scheme_name = scheme_name
    
    def resolve_steps(self, top_node, steps):
        """Resolve path from top_node, return node found or None.
        Steps are in order top_node to node.
        """
        pass

    def create_steps(self, top_node, node):
        """Return list of steps from top_node to node.
        Steps returned are in reverse order.
        """
        pass
    
class ChildNodePathScheme(BaseNodePathScheme):
    def __init__(self):
        BaseNodePathScheme.__init__(self, 'child')
        
    def resolve_steps(self, top_node, steps):
        node = top_node
        for step in steps:
            node = node.childNodes.item(int(step))
            if node is None:
                return None
        return node # return node

    def create_steps(self, top_node, node):
        steps = []
        # FIXME: is it safe to compare nodes like this?
        while node is not top_node:
            parent = node.parentNode
            if parent is None:
                break
            # FIXME: can index() be used in all python DOMs?
            steps.append(str(parent.childNodes.index(node)))
            node = parent
        steps.reverse()
        return steps

# create the registry
registry = NodePathSchemeRegistry()



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Facility to use the Expat parser to load a ParsedXML.DOM instance
from a string or file."""

# Warning!
#
# This module is tightly bound to the implementation details of the
# Parsed XML DOM and can't be used with other DOM implementations.  This
# is due, in part, to a lack of appropriate methods in the DOM (there is
# no way to create Entity and Notation nodes via the DOM Level 2
# interface), and for performance.  The later is the cause of some fairly
# cryptic code.
#
# Performance hacks:
#
#   -  .character_data_handler() has an extra case in which continuing
#      data is appended to an existing Text node; this can be a
#      substantial speedup since Expat seems to break data at every
#      newline.
#
#   -  Determining that a node exists is done using an identity comparison
#      with None rather than a truth test; this avoids searching for and
#      calling any methods on the node object if it exists.  (A rather
#      nice speedup is achieved this way as well!)

import string

import Core
import XMLExtended

from xml.parsers import expat


class Options:
    """Features object that has variables set for each DOMBuilder feature.

    The DOMBuilder class uses an instance of this class to pass settings to
    the ExpatBuilder class.
    """

    # Note that the DOMBuilder class in LoadSave constrains which of these
    # values can be set using the DOM Level 3 LoadSave feature.

    namespaces = 1
    namespace_declarations = 1
    validation = 0
    external_general_entities = 1
    external_parameter_entities = 1
    validate_if_cm = 0
    create_entity_ref_nodes = 1
    entity_nodes = 1
    white_space_in_element_content = 1
    cdata_nodes = 1
    comments = 1
    charset_overrides_xml_encoding = 1

    errorHandler = None
    filter = None


class ExpatBuilder:
    """Document builder that uses Expat to build a ParsedXML.DOM document
    instance."""

    def __init__(self, options=None):
        if options is None:
            options = Options()
        self._options = options
        self._parser = None
        self.reset()

    try:
        {}.setdefault
    except AttributeError:
        def _intern(self, s):
            try:
                return self._interns[s]
            except KeyError:
                self._interns[s] = s
                return s
    else:
        def _intern(self, s):
            return self._interns.setdefault(s, s)

    def createParser(self):
        """Create a new parser object."""
        return expat.ParserCreate()

    def getParser(self):
        """Return the parser object, creating a new one if needed."""
        if not self._parser:
            self._parser = self.createParser()
            self.install(self._parser)
        return self._parser

    def reset(self):
        """Free all data structures used during DOM construction."""
        self.document = None
        self._cdata = 0
        self._standalone = -1
        self._version = None
        self._encoding = None
        self._doctype_args = None
        self._entities = []
        self._notations = []
        self._pre_doc_events = []
        self._attr_info = {}
        self._elem_info = {}
        self._interns = {}

    def install(self, parser):
        """Install the callbacks needed to build the DOM into the parser."""
        # This creates circular references!
        parser.StartDoctypeDeclHandler = self.start_doctype_decl_handler
        parser.StartElementHandler = self.start_element_handler
        parser.EndElementHandler = self.end_element_handler
        parser.ProcessingInstructionHandler = self.pi_handler
        parser.CharacterDataHandler = self.character_data_handler
        parser.EntityDeclHandler = self.entity_decl_handler
        parser.NotationDeclHandler = self.notation_decl_handler
        parser.CommentHandler = self.comment_handler
        parser.StartCdataSectionHandler = self.start_cdata_section_handler
        parser.EndCdataSectionHandler = self.end_cdata_section_handler
        parser.ExternalEntityRefHandler = self.external_entity_ref_handler
        parser.ordered_attributes = 1
        parser.specified_attributes = 1
        parser.XmlDeclHandler = self.xml_decl_handler
        parser.ElementDeclHandler = self.element_decl_handler
        parser.AttlistDeclHandler = self.attlist_decl_handler

    def parseFile(self, file):
        """Parse a document from a file object, returning the document
        node."""
        parser = self.getParser()
        first_buffer = 1
        strip_newline = 0
        while 1:
            buffer = file.read(16*1024)
            if not buffer:
                break
            if strip_newline:
                if buffer[0] == "\n":
                    buffer = buffer[1:]
                strip_newline = 0
            if buffer and buffer[-1] == "\r":
                strip_newline = 1
            buffer = _normalize_lines(buffer)
            parser.Parse(buffer, 0)
            if first_buffer and self.document:
                if self.document.doctype:
                    self._setup_subset(buffer)
                first_buffer = 0
        parser.Parse("", 1)
        doc = self.document
        self.reset()
        self._parser = None
        return doc

    def parseString(self, string):
        """Parse a document from a string, returning the document node."""
        string = _normalize_lines(string)
        parser = self.getParser()
        parser.Parse(string, 1)
        self._setup_subset(string)
        doc = self.document
        self.reset()
        self._parser = None
        return doc

    def _setup_subset(self, buffer):
        """Load the internal subset if there might be one."""
        if self.document.doctype:
            extractor = InternalSubsetExtractor()
            extractor.parseString(buffer)
            subset = extractor.getSubset()
            if subset is not None:
                d = self.document.doctype.__dict__
                d['internalSubset'] = subset

    def start_doctype_decl_handler(self, doctypeName, systemId, publicId,
                                   has_internal_subset):
        self._pre_doc_events.append(("doctype",))
        self._doctype_args = (self._intern(doctypeName), publicId, systemId)

    def pi_handler(self, target, data):
        target = self._intern(target)
        if self.document is None:
            self._pre_doc_events.append(("pi", target, data))
        else:
            node = self.document.createProcessingInstruction(target, data)
            self.curNode.appendChild(node)

    def character_data_handler(self, data):
        if self._cdata:
            if (self._cdata_continue
                and (self.curNode._children[-1].nodeType
                     == Core.Node.CDATA_SECTION_NODE)):
                d = self.curNode._children[-1].__dict__
                data = d['data'] + data
                d['data'] = d['nodeValue'] = data
                return
            node = self.document.createCDATASection(data)
            self._cdata_continue = 1
        elif (self.curNode._children
              and self.curNode._children[-1].nodeType == Core.Node.TEXT_NODE):
            node = self.curNode._children[-1]
            data = node.data + data
            d = node.__dict__
            d['data'] = d['nodeValue'] = data
            return
        else:
            node = self.document.createTextNode(data)
        self.curNode.appendChild(node)

    def entity_decl_handler(self, entityName, is_parameter_entity, value,
                            base, systemId, publicId, notationName):
        if is_parameter_entity:
            # we don't care about parameter entities for the DOM
            return
        if not self._options.entity_nodes:
            return
        entityName = self._intern(entityName)
        notationName = self._intern(notationName)
        node = XMLExtended.Entity(entityName, publicId, systemId, notationName)
        if value is not None:
            # internal entity
            child = Core.Text(value)
            # must still get to parent, even if entity isn't _in_tree
            child.__dict__['_in_tree'] = 1
            child.__dict__['_readonly'] = 1
            node.__dict__['_children'] = [child]
        self._entities.append(node)

    def notation_decl_handler(self, notationName, base, systemId, publicId):
        notationName = self._intern(notationName)
        node = XMLExtended.Notation(notationName, publicId, systemId)
        self._notations.append(node)

    def comment_handler(self, data):
        if self._options.comments:
            if self.document is None:
                self._pre_doc_events.append(("comment", data))
            else:
                node = self.document.createComment(data)
                self.curNode.appendChild(node)

    def start_cdata_section_handler(self):
        if self._options.cdata_nodes:
            self._cdata = 1
            self._cdata_continue = 0

    def end_cdata_section_handler(self):
        self._cdata = 0
        self._cdata_continue = 0

    def external_entity_ref_handler(self, context, base, systemId, publicId):
        return 1

    def start_element_handler(self, name, attributes):
        name = self._intern(name)
        if self.document is None:
            doctype = self._create_doctype()
            doc = Core.theDOMImplementation.createDocument(
                None, name, doctype)
            if self._standalone >= 0:
                doc.standalone = self._standalone
            doc.encoding = self._encoding
            doc.version = self._version
            doc.__dict__['_elem_info'] = self._elem_info
            doc.__dict__['_attr_info'] = self._attr_info
            self.document = doc
            self._include_early_events()
            node = doc.documentElement
            # chicken & egg: if this isn't inserted here, the document
            # element doesn't get the information about the defined
            # attributes for its element type
            if self._attr_info.has_key(name):
                node.__dict__['_attr_info'] = self._attr_info[name]
        else:
            node = self.document.createElement(name)
            self.curNode.appendChild(node)
        self.curNode = node

        if attributes:
            L = []
            for i in range(0, len(attributes), 2):
                L.append([None, self._intern(attributes[i]),
                          None, None, attributes[i+1], 1])
            node.__dict__['_attributes'] = L

    def end_element_handler(self, name):
        curNode = self.curNode
        assert curNode.tagName == name, "element stack messed up!"
        self.curNode = self.curNode.parentNode
        self._handle_white_text_nodes(curNode)
        if self._options.filter:
            self._options.filter.endElement(curNode)

    def _handle_white_text_nodes(self, node):
        info = self._elem_info.get(node.tagName)
        if not info:
            return
        type = info[0]
        if type in (expat.model.XML_CTYPE_ANY,
                    expat.model.XML_CTYPE_MIXED):
            return
        #
        # We have element type information; look for text nodes which
        # contain only whitespace.
        #
        L = []
        for child in node.childNodes:
            if (  child.nodeType == Core.Node.TEXT_NODE
                  and not string.strip(child.data)):
                L.append(child)
        #
        # Depending on the options, either mark the nodes as ignorable
        # whitespace or remove them from the tree.
        #
        for child in L:
            if self._options.white_space_in_element_content:
                child.__dict__['isWhitespaceInElementContent'] = 1
            else:
                node.removeChild(child)

    def element_decl_handler(self, name, model):
        self._elem_info[self._intern(name)] = model

    def attlist_decl_handler(self, elem, name, type, default, required):
        elem = self._intern(elem)
        name = self._intern(name)
        type = self._intern(type)
        if self._attr_info.has_key(elem):
            L = self._attr_info[elem]
        else:
            L = []
            self._attr_info[elem] = L
        L.append([None, name, None, None, default, 0, type, required])

    def xml_decl_handler(self, version, encoding, standalone):
        self._version = version
        self._encoding = encoding
        self._standalone = standalone

    def _create_doctype(self):
        if not self._doctype_args:
            return
        doctype = apply(Core.theDOMImplementation.createDocumentType,
                        self._doctype_args)
        doctype._entities[:] = self._entities
        self._entities = doctype._entities
        doctype._notations[:] = self._notations
        self._notations = doctype._notations
        return doctype

    def _include_early_events(self):
        doc = self.document
        if self._doctype_args:
            docelem = doc.doctype
        else:
            docelem = doc.documentElement
        for event in self._pre_doc_events:
            t = event[0]
            if t == "comment":
                node = doc.createComment(event[1])
            elif t == "doctype":
                # marker; switch to before docelem
                docelem = doc.documentElement
                continue
            elif t == "pi":
                node = doc.createProcessingInstruction(event[1], event[2])
            else:
                raise RuntimeError, "unexpected early event type: " + `t`
            doc.insertBefore(node, docelem)


def _normalize_lines(s):
    """Return a copy of 's' with line-endings normalized according to
    XML 1.0 section 2.11."""
    s = string.replace(s, "\r\n", "\n")
    return string.replace(s, "\r", "\n")


# framework document used by the fragment builder.
# Takes a string for the doctype, subset string, and namespace attrs string.

_FRAGMENT_BUILDER_INTERNAL_SYSTEM_ID = \
    "http://xml.zope.org/entities/fragment-builder/internal"

_FRAGMENT_BUILDER_TEMPLATE = (
    '''\
<!DOCTYPE wrapper
  %%s [
  <!ENTITY fragment-builder-internal
    SYSTEM "%s">
%%s
]>
<wrapper %%s
>&fragment-builder-internal;</wrapper>'''
    % _FRAGMENT_BUILDER_INTERNAL_SYSTEM_ID)


class FragmentBuilder(ExpatBuilder):
    """Builder which constructs document fragments given XML source
    text and a context node.

    The context node is expected to provide information about the
    namespace declarations which are in scope at the start of the
    fragment.
    """

    def __init__(self, context, options=None):
        if context.nodeType == Core.Node.DOCUMENT_NODE:
            self.originalDocument = context
            self.context = context
        else:
            self.originalDocument = context.ownerDocument
            self.context = context
        ExpatBuilder.__init__(self, options)

    def reset(self):
        ExpatBuilder.reset(self)
        self.fragment = None

    def parseFile(self, file):
        """Parse a document fragment from a file object, returning the
        fragment node."""
        return self.parseString(file.read())

    def parseString(self, string):
        """Parse a document fragment from a string, returning the
        fragment node."""
        self._source = string
        parser = self.getParser()
        doctype = self.originalDocument.doctype
        ident = ""
        if doctype:
            subset = doctype.internalSubset or self._getDeclarations()
            if doctype.publicId:
                ident = ('PUBLIC "%s" "%s"'
                         % (doctype.publicId, doctype.systemId))
            elif doctype.systemId:
                ident = 'SYSTEM "%s"' % doctype.systemId
        else:
            subset = ""
        nsattrs = self._getNSattrs() # get ns decls from node's ancestors
        document = _FRAGMENT_BUILDER_TEMPLATE % (ident, subset, nsattrs)
        try:
            parser.Parse(document, 1)
        except:
            self.reset()
            raise
        fragment = self.fragment
        self.reset()
##         self._parser = None
        return fragment

    def _getDeclarations(self):
        """Re-create the internal subset from the DocumentType node.

        This is only needed if we don't already have the
        internalSubset as a string.
        """
        doctype = self.originalDocument.doctype
        s = ""
        if doctype:
            for i in range(doctype.notations.length):
                notation = doctype.notations.item(i)
                if s:
                    s = s + "\n  "
                s = "%s<!NOTATION %s" % (s, notation.nodeName)
                if notation.publicId:
                    s = '%s PUBLIC "%s"\n             "%s">' \
                        % (s, notation.publicId, notation.systemId)
                else:
                    s = '%s SYSTEM "%s">' % (s, notation.systemId)
            for i in range(doctype.entities.length):
                entity = doctype.entities.item(i)
                if s:
                    s = s + "\n  "
                s = "%s<!ENTITY %s" % (s, entity.nodeName)
                if entity.publicId:
                    s = '%s PUBLIC "%s"\n             "%s"' \
                        % (s, entity.publicId, entity.systemId)
                elif entity.systemId:
                    s = '%s SYSTEM "%s"' % (s, entity.systemId)
                else:
                    s = '%s "%s"' % (s, entity.firstChild.data)
                if entity.notationName:
                    s = "%s NOTATION %s" % (s, entity.notationName)
                s = s + ">"
        return s

    def _getNSattrs(self):
        return ""

    def external_entity_ref_handler(self, context, base, systemId, publicId):
        if systemId == _FRAGMENT_BUILDER_INTERNAL_SYSTEM_ID:
            # this entref is the one that we made to put the subtree
            # in; all of our given input is parsed in here.
            old_document = self.document
            old_cur_node = self.curNode
            self._save_namespace_decls()
            parser = self._parser.ExternalEntityParserCreate(context)
            self._restore_namespace_decls()
            # put the real document back, parse into the fragment to return
            self.document = self.originalDocument
            self.fragment = self.document.createDocumentFragment()
            self.curNode = self.fragment
            try:
                parser.Parse(self._source, 1)
            finally:
                self.curNode = old_cur_node
                self.document = old_document
                self._source = None
            return -1
        else:
            return ExpatBuilder.external_entity_ref_handler(
                self, context, base, systemId, publicId)

    def _save_namespace_decls(self):
        pass

    def _restore_namespace_decls(self):
        pass


class Namespaces:
    """Mix-in class for builders; adds support for namespaces."""

    def _initNamespaces(self):
        #
        # These first two dictionaries are used to track internal
        # namespace state, and contain all "current" declarations.
        #
        # URI -> [prefix, prefix, prefix...]
        #
        # The last prefix in list is most recently declared; there's
        # no way to be sure we're using the right one if more than one
        # has been defined for a particular URI.
        self._nsmap = {
            Core.XML_NS: ["xml"],
            Core.XMLNS_NS: ["xmlns"],
            }
        # prefix -> URI
        self._prefixmap = {
            "xml": [Core.XML_NS],
            "xmlns": [Core.XMLNS_NS],
            }
        #
        # These dictionaries are used to store the namespace
        # declaractions made on a single element; they are used to add
        # the attributes of the same name to the DOM structure.  When
        # added to the DOM, they are replaced with new, empty
        # dictionaries on the Builder object.
        #
        self._ns_prefix_uri = {}
        self._ns_uri_prefixes = {}
        # list of (prefix, uri) ns declarations.  Namespace attrs are
        # constructed from this and added to the element's attrs.
        self._ns_ordered_prefixes = []

    def createParser(self):
        """Create a new namespace-handling parser."""
        return expat.ParserCreate(namespace_separator=" ")

    def install(self, parser):
        """Insert the namespace-handlers onto the parser."""
        ExpatBuilder.install(self, parser)
        parser.StartNamespaceDeclHandler = self.start_namespace_decl_handler
        parser.EndNamespaceDeclHandler = self.end_namespace_decl_handler

    def start_namespace_decl_handler(self, prefix, uri):
        "push this namespace declaration on our storage"
        #
        # These are what we use internally:
        #
        prefix = self._intern(prefix)
        uri = self._intern(uri)
        L = self._nsmap.get(uri)
        if L is None:
            self._nsmap[uri] = L = []
        L.append(prefix)
        L = self._prefixmap.get(prefix)
        if L is None:
            self._prefixmap[prefix] = L = []
        L.append(uri)
        #
        # These are used to provide namespace declaration info to the DOM:
        #
        self._ns_prefix_uri[prefix] = uri
        L = self._ns_uri_prefixes.get(uri)
        if not L:
            self._ns_uri_prefixes[uri] = L = []
        L.append(prefix)
        self._ns_ordered_prefixes.append((prefix, uri))

    def end_namespace_decl_handler(self, prefix):
        "pop the latest namespace declaration."
        uri = self._prefixmap[prefix].pop()
        self._nsmap[uri].pop()

    def _save_namespace_decls(self):
        """Save the stored namespace decls and reset the new ones.
        This lets us launch another parser and have its namespace declarations
        not affect future elements.  Must be called outside of any start/end
        namespace_decl_handler calls."""
        self._oldnsmap = self._nsmap
        self._oldprefixmap = self._prefixmap
        self._oldns_prefix_uri = self._ns_prefix_uri
        self._oldns_uri_prefixes = self._ns_uri_prefixes
        self._oldns_ordered_prefixes = self._ns_ordered_prefixes
        self._initNamespaces()

    def _restore_namespace_decls(self):
        "Restore the namespace decls from _save_namespace_decls."
        self._nsmap = self._oldnsmap
        self._prefixmap = self._oldprefixmap
        self._ns_prefix_uri = self._oldns_prefix_uri
        self._ns_uri_prefixes = self._oldns_uri_prefixes
        self._ns_ordered_prefixes = self._oldns_ordered_prefixes

    def start_element_handler(self, name, attributes):
        if ' ' in name:
            uri, localname = string.split(name, ' ')
            localname = self._intern(localname)
            uri = self._intern(uri)
            prefix = self._intern(self._nsmap[uri][-1])
            if prefix:
                qname = "%s:%s" % (prefix, localname)
            else:
                qname = localname
        else:
            uri = None
            qname = name
            localname = prefix = None
        qname = self._intern(qname)
        if self.document is None:
            doctype = self._create_doctype()
            doc = Core.theDOMImplementation.createDocument(
                uri, qname, doctype)
            if self._standalone >= 0:
                doc.standalone = self._standalone
            doc.encoding = self._encoding
            doc.version = self._version
            doc.__dict__['_elem_info'] = self._elem_info
            doc.__dict__['_attr_info'] = self._attr_info
            self.document = doc
            self._include_early_events()
            node = doc.documentElement
            # chicken & egg: if this isn't inserted here, the document
            # element doesn't get the information about the defined
            # attributes for its element type
            if self._attr_info.has_key(qname):
                node.__dict__['_attr_info'] = self._attr_info[qname]
        else:
            node = self.document.createElementNS(
                uri, qname, (prefix, localname))
            self.curNode.appendChild(node)
        self.curNode = node

        L = [] # [[namespaceURI, qualifiedName, localName, prefix,
               #   value, specified]]
        if self._ns_ordered_prefixes and self._options.namespace_declarations:
            for prefix, uri in self._ns_ordered_prefixes:
                if prefix:
                    attrPrefix = "xmlns"
                    tagName = self._intern('%s:%s' % (attrPrefix, prefix))
                else:
                    attrPrefix = tagName = "xmlns"
                L.append([Core.XMLNS_NS, tagName, self._intern(prefix),
                          attrPrefix, uri, 1])
        if attributes:
            # This uses the most-recently declared prefix, not necessarily
            # the right one.
            for i in range(0, len(attributes), 2):
                aname = attributes[i]
                value = attributes[i+1]
                if ' ' in aname:
                    uri, localname = string.split(aname, ' ')
                    localname = self._intern(localname)
                    prefix = self._intern(self._nsmap[uri][-1])
                    uri = self._intern(uri)
                    if prefix:
                        qualifiedname = self._intern(
                            '%s:%s' % (prefix, localname))
                    else:
                        qualifiedname = localname
                    L.append([uri, qualifiedname, localname, prefix, value, 1])
                else:
                    name = self._intern(aname)
                    L.append([None, name, name, None, value, 1])
        if L:
            node.__dict__['_attributes'] = L

        if self._ns_prefix_uri:
            # insert this stuff on the element:
            d = node.__dict__
            d['_ns_prefix_uri'] = self._ns_prefix_uri
            d['_ns_uri_prefixes'] = self._ns_uri_prefixes
            # reset for the next:
            self._ns_prefix_uri = {}
            self._ns_uri_prefixes = {}
            self._ns_ordered_prefixes = []

    def end_element_handler(self, name):
        if ' ' in name:
            uri, localname = string.split(name, ' ')
            assert (self.curNode.namespaceURI == uri
                    and self.curNode.localName == localname), \
                    "element stack messed up! (namespace)"
        else:
            assert self.curNode.nodeName == name, \
                   "element stack messed up - bad nodeName"
            assert self.curNode.namespaceURI is None, \
                   "element stack messed up - bad namespaceURI"
        self._handle_white_text_nodes(self.curNode)
        self.curNode = self.curNode.parentNode


class ExpatBuilderNS(Namespaces, ExpatBuilder):
    """Document builder that supports namespaces."""

    def reset(self):
        ExpatBuilder.reset(self)
        self._initNamespaces()


class FragmentBuilderNS(Namespaces, FragmentBuilder):
    """Fragment builder that supports namespaces."""

    def reset(self):
        FragmentBuilder.reset(self)
        self._initNamespaces()

    def _getNSattrs(self):
        """Return string of namespace attributes from this element and
        ancestors."""
        attrs = ""
        context = self.context
        L = []
        while context:
            if hasattr(context, '_ns_prefix_uri'):
                for prefix, uri in context._ns_prefix_uri.items():
                    # add every new NS decl from context to L and attrs string
                    if prefix in L:
                        continue
                    L.append(prefix)
                    if prefix:
                        declname = "xmlns:" + prefix
                    else:
                        declname = "xmlns"
                    if attrs:
                        attrs = "%s\n    %s='%s'" % (attrs, declname, uri)
                    else:
                        attrs = " %s='%s'" % (declname, uri)
            context = context.parentNode
        return attrs


class ParseEscape(Exception):
    """Exception raised to short-circuit parsing in InternalSubsetExtractor."""
    pass

class InternalSubsetExtractor(ExpatBuilder):
    """XML processor which can rip out the internal document type subset."""

    def getSubset(self):
        """Return the internal subset as a string."""
        subset = self.subset
        while subset and subset[0] != "[":
            del subset[0]
        if subset:
            x = subset.index("]")
            return string.join(subset[1:x], "")
        else:
            return None

    def parseFile(self, file):
        try:
            ExpatBuilder.parseFile(self, file)
        except ParseEscape:
            pass

    def parseString(self, string):
        try:
            ExpatBuilder.parseString(self, string)
        except ParseEscape:
            pass

    def install(self, parser):
        parser.StartDoctypeDeclHandler = self.start_doctype_decl_handler
        parser.EndDoctypeDeclHandler = self.end_doctype_decl_handler
        parser.StartElementHandler = self.start_element_handler

    def start_doctype_decl_handler(self, *args):
        self.subset = []
        self.getParser().DefaultHandler = self.default_handler

    def end_doctype_decl_handler(self):
        self.getParser().DefaultHandler = None
        raise ParseEscape()

    def start_element_handler(self, name, attrs):
        raise ParseEscape()

    def default_handler(self, s):
        self.subset.append(s)


def parse(file, namespaces=1):
    """Parse a document, returning the resulting Document node.

    'file' may be either a file name or an open file object.
    """
    if namespaces:
        builder = ExpatBuilderNS()
    else:
        builder = ExpatBuilder()

    if isinstance(file, type('')):
        fp = open(file, 'rb')
        result = builder.parseFile(fp)
        fp.close()
    else:
        result = builder.parseFile(file)
    return result


def parseFragment(file, context, namespaces=1):
    """Parse a fragment of a document, given the context from which it was
    originally extracted.  context should be the parent of the node(s) which
    are in the fragment.

    'file' may be either a file name or an open file object.
    """
    if namespaces:
        builder = FragmentBuilderNS(context)
    else:
        builder = FragmentBuilder(context)

    if isinstance(file, type('')):
        fp = open(file, 'rb')
        result = builder.parseFile(fp)
        fp.close()
    else:
        result = builder.parseFile(file)
    return result


def makeBuilder(options):
    """Create a builder based on an Options object."""
    if options.namespaces:
        return ExpatBuilderNS(options)
    else:
        return ExpatBuilder(options)



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"test that the ParsedXML DOM object acquire implicitly"

import os, sys
  
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Testing import ZopeTestCase
  
ZopeTestCase.installProduct('ParsedXML')

from Products.ParsedXML import ParsedXML

# We currently have to have a Zope instance
# to mount; we need a persistent container & Zope traversal to acquire
# properly, because we don't want to acquire through the transient
# proxy objects.

def checkAcquire(aqer, aqee, name):
    "check that acquisition of name from aqee to aqer happens"
    assert hasattr(aqer, name)
    assert getattr(aqer, name) == getattr(aqee, name)
    assert getattr(aqer, name) is getattr(aqee, name)

def checkAcquireNot(aqer, aqee, name):
    "check that acquisition of name from aqee to aqer doesn't happen"
    assert getattr(aqer, name) != getattr(aqee, name)
    assert getattr(aqer, name) is not getattr(aqee, name)


class WrappedAcquisitionTestCase(ZopeTestCase.ZopeTestCase):

    def afterSetUp(self):
        
        self.tmpId = 'TempParsedXMLUnitTestInstance'
        self.app._setObject(self.tmpId, ParsedXML.ParsedXML(self.tmpId))
        self.doc = getattr(self.app, self.tmpId)
        
        self.app.string = 'a string'
        self.app.firstChild = "a string that shouldn't be acquired"

        elt1 = self.doc.createElement('eggs')
        elt2 = self.doc.createElement('ham')
        self.doc.documentElement.appendChild(elt1)
        self.doc.documentElement.appendChild(elt2)
        self.doc.documentElement.setAttribute('color', 'green')

    def _acquisitionTest(self, DOMObj):
        checkAcquire(DOMObj, self.app, 'string')
        checkAcquireNot(DOMObj, self.app, 'firstChild')

    def testTraversalAcquisition(self):
        "make sure we can acquire through DOM traversal"
        self._acquisitionTest(self.doc)
        self._acquisitionTest(self.doc.documentElement)
        self._acquisitionTest(self.doc.documentElement.firstChild)
        self._acquisitionTest(self.doc.documentElement.firstChild.nextSibling)
        self._acquisitionTest(self.doc.documentElement.ownerDocument)
        self._acquisitionTest(self.doc.documentElement.getAttributeNode(
            'color'))

    def testMethodAcquisition(self):
        "make sure we acquire through proxied DOM methods"
        elt = self.doc.createElement('foo')
        self._acquisitionTest(self.doc.documentElement.appendChild(elt))

    def testNodeListAcquisition(self):
        "make sure we can acquire through a NodeList"
        self._acquisitionTest(self.doc.documentElement.childNodes.item(0))

    def testNamedNodeMapAcquisition(self):
        "make sure we can acquire through a NamedNodeMap"
        self._acquisitionTest(
            self.doc.documentElement.attributes.getNamedItem('color'))

if __name__ == '__main__':
    framework()
else:
    import unittest
    def test_suite():
        suite = unittest.TestSuite()
 
        suite.addTest(unittest.makeSuite(WrappedAcquisitionTestCase))
        
        return suite




##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Digital Creation's light-weight, acquisition-based DOM implementation."""

__version__ = '$Revision: 1.5 $'

# Work around the possible lack of a decent XML package; the
# Exceptions module will masquerade as xml.dom if it needs to.
# See the comments in Exceptions for a full explanation.
#
import Exceptions

from Core import theDOMImplementation



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""
Non-DOM helper functions useful when working with either of our DOM
implementations, but not specific to any instantiation.
"""

from xml.dom import Node

import DOM.ExpatBuilder
import PrettyPrinter
from StringIO import StringIO


def parseFile(node, file, namespaces = 1):
    """
    Parse XML file, replace node with the resulting tree,
    return replacement.
    Node must be in an existing DOM tree if not a Document.
    """
    if node.nodeType == Node.DOCUMENT_NODE:
        return DOM.ExpatBuilder.parse(file, namespaces)
    elif node.nodeType == Node.DOCUMENT_FRAGMENT_NODE:
        raise Exception, "replacing a document fragment doesn't make sense"
    else:
        fragment = DOM.ExpatBuilder.parseFragment(
            file, node.parentNode, namespaces)
        #if fragment.childNodes.length != 1:
        #    # we could do this, actually, if we wanted
        #    raise Exception, "replacing a Node with less or more " + \
        #          "than one tree of Nodes is not implemented"
        sib = node.nextSibling
        parent = node.parentNode
        parent.removeChild(node)
        #print fragment.childNodes.length
        #for child in fragment.childNodes:
        #    print child.nodeType, node.nodeName
        #    if child.nodeType == Node.TEXT_NODE:
        #        print repr(child.data) 
        return parent.insertBefore(fragment.firstChild, sib) # frag.firstChild
         

def writeStream(node, stream = None, encoding=None, prettyPrint = 0):
    "Write the XML representation of node to stream."
    if stream is None:
        stream = StringIO()
    PrettyPrinter.PrintVisitor(
        node, stream, encoding, prettyPrint=prettyPrint)()
    return stream

    



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Test that checks the fragility of using acquistion wrappers to
indicate tree hierarchy.

The current implementation of the DOM uses acquisition wrappers to
store references to a node's parent node.  While this allows very
efficient access to the parent, it is fragile in the case of multiple
wrappers referring to the same node.  If client code holds two
wrappers for a node and modifies the node's position in the tree using
one of them, the other will store incorrect information on the shape
of the tree.

For acquisition to be used to adequately be used to present the
containment hierarchy for a node, a complete chain of wrappers would
need to be constructed each time a node is reparented, starting from
the outermost node in the ancestor chain.  Application code would need
to be wary that old references to a node are replaced by the new
wrapper.  The Python DOM API makes no such requirement at the present
time, nor should it need to.

This script shows a contrived example that exercises this DOM bug.
While this particular code is unlikely in real applications, the ease
with which multiple acquisition wrappers can be produced in more
complex application code is easy to see.

"""
import os, sys
  
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Testing import ZopeTestCase
  
ZopeTestCase.installProduct('ParsedXML')

from Products.ParsedXML.DOM.ExpatBuilder import ExpatBuilder

class AcquisitionPainTestCase(ZopeTestCase.ZopeTestCase):

    def afterSetUp(self):
        self.doc = ExpatBuilder().parseString("<doc><e1/><e2/></doc>")

    def testParentReferenceIntegrity(self):
        e1a = self.doc.documentElement.firstChild
        e1b = self.doc.documentElement.firstChild
        e2 = e1b.nextSibling
        e2.appendChild(e1b)
        assert e1a.parentNode.isSameNode(e1b.parentNode), \
               "Two references to the same node return different parent nodes."

if __name__ == '__main__':
    framework()
else:
    import unittest
    def test_suite():
        suite = unittest.TestSuite()
        suite.addTest(unittest.makeSuite(AcquisitionPainTestCase))
        return suite



from setuptools import setup, find_packages
import os

version = '1.5'

setup(name='Products.ParsedXML',
      version=version,
      description="Parsed XML allows you to use XML objects in the Zope 2 environment.",
      long_description=open(os.path.join("Products", "ParsedXML", "README.txt")).read() + "\n" +
                       open(os.path.join("Products", "ParsedXML", "CREDITS.txt")).read() + "\n" +
                       open(os.path.join("Products", "ParsedXML", "CHANGES.txt")).read(),
      # Get more strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
              "Framework :: Zope2",
              "License :: OSI Approved :: Zope Public License",
              "Programming Language :: Python",
              "Topic :: Software Development :: Libraries :: Python Modules",
              "Topic :: Text Processing :: Markup :: XML",
              ],
      keywords='parsedxml xml zope2',
      author='Zope community, and various others contributors',
      author_email='info@infrae.com',
      url='http://www.zope.org/Members/faassen/ParsedXML',
      license='ZPL 2.0 and Fourthought license',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['Products'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
              'setuptools',
              ],
      )



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import re
from DOM.Core import Node, XMLNS_NS, XML_NS
import sys
from StringIO import StringIO

# an XML printer which:
#   * can do pretty printing, optionally
#   * as opposed to the one defined in Printer.py, actually should
#     be less buggy

class PrintVisitor:
    
    def __init__(self, root, stream=sys.stdout, encoding=None,
                 entityReferenceExpansion=1,
                 prettyPrint=0, indentLevel=2):

        self.root = root
        self.stream = stream
        self.encoding = encoding
        self.entityReferenceExpansion = entityReferenceExpansion
        self.prettyPrint = prettyPrint
        self.indent = 0
        self.indentLevel = indentLevel
        
        self.nodeType2method = { 
            Node.ELEMENT_NODE: self.renderElement,
            Node.ATTRIBUTE_NODE: self.renderAttr,
            Node.TEXT_NODE: self.renderText,
            Node.CDATA_SECTION_NODE: self.renderCDATASection,
            Node.ENTITY_REFERENCE_NODE: self.renderEntityReference,
            Node.ENTITY_NODE: self.renderEntity,
            Node.PROCESSING_INSTRUCTION_NODE:\
            self.renderProcessingInstruction,
            Node.COMMENT_NODE: self.renderComment,
            Node.DOCUMENT_NODE: self.renderDocument,
            Node.DOCUMENT_TYPE_NODE: self.renderDocumentType,
            Node.DOCUMENT_FRAGMENT_NODE: self.renderDocumentFragment,
            Node.NOTATION_NODE: self.renderNotation,
        }

    def renderAll(self):
        return self.render(self.stream, self.root)

    __call__ = renderAll
    
    def render(self, f, node):
        self.nodeType2method[node.nodeType](f, node)
        
    def renderElement(self, f, node):
        if self.prettyPrint:
            f.write(" " * self.indent * self.indentLevel)
        f.write("<")
        f.write(node.tagName)
        for attribute in node.attributes.values():
            self.renderAttr(f, attribute)
        if not node.hasChildNodes():
            f.write('/>')
            if self.prettyPrint:
                f.write("\n")
        else:
            f.write('>')
            prettyPrint = self.prettyPrint

            stream = f
            if prettyPrint:
                f.write("\n")                    
                no_indentation = 0
                for child in node.childNodes:
                    if (child.nodeType == Node.TEXT_NODE and
                        child.data.strip() != ''):
                        no_indentation = 1
                        break
                if no_indentation:
                    stream = StringIO()
                    self.prettyPrint = 0
                self.indent += 1
                
            for child in node.childNodes:
                self.render(stream, child)

            self.prettyPrint = prettyPrint
            
            if prettyPrint:
                if no_indentation:
                    f.write(indentBlock(
                        stream.getvalue().strip(),
                        self.indent * self.indentLevel, 70))
                    f.write('\n')
                self.indent -= 1
                
                f.write(" " * self.indent * self.indentLevel)
            f.write("</%s>" % node.tagName)
            if self.prettyPrint:
                f.write("\n")

    def renderAttr(self, f, node):
        if not node.specified:
            return
        text, delimiter = _translateCdataAttr(node.value)
        f.write(" %s=%s%s%s" % (node.name,
                                delimiter, text, delimiter))
        
    def renderText(self, f, node):
        data = node.data
        if self.prettyPrint:
            data = node.data.strip()
            if data == "":
                return
            data = indentBlock(data, self.indent * self.indentLevel, 70)
        f.write(_translateCdata(data))
        if self.prettyPrint:
            f.write('\n')
            
    def renderCDATASection(self, f, node):
        f.write("<![CDATA[")
        f.write(node.data.replace("]]>", "]]]><![CDATA[]>"))
        f.write("]]>")

    def renderEntityReference(self, f, node):
        f.write('&')
        f.write(node.nodeName)
        f.write(';')

    def renderEntity(self, f, node):
        st = "<!ENTITY " + node.nodeName
        if not node.systemId:
            # internal entity
            s = node.firstChild.data
            st = '%s "%s"' % (st, _translateCdata(s))
        if node.publicId:
            st = st + ' PUBLIC "%s"' % node.publicId
            if node.systemId:
                st = '%s "%s"' % (st, node.systemId)
        elif node.systemId:
            st = st + ' SYSTEM "%s"' % node.systemId
        if node.notationName:
            st = st + ' NDATA %s' % node.notationName
        f.write(st + '>\n')
    
    def renderProcessingInstruction(self, f, node):
        f.write('<?')
        f.write(node.target + ' ')
        f.write(node.data)
        f.write('?>')

    def renderComment(self, f, node):
        f.write('<!--')
        f.write(node.data)
        f.write('-->')

    def renderDocument(self, f, node):    
        f.write('<?xml version="1.0" ?>\n')
        #if self.encoding:
        #    f.write(' encoding="%s"' % self.encoding)
        #f.write(' ?>\n')
        for child in node.childNodes:
            self.render(f, child)
        f.write('\n')
        
    def renderDocumentType(self, f, node):
        if (not node.entities.length and
            not node.notations.length and
            not node.systemId):
            return

        f.write("<!DOCTYPE ")
        f.write(node.name)

        if node.systemId:
            if node.publicId:
                if '"' not in node.publicId:
                    f.write(' PUBLIC "' + node.publicId + '" ')
                else:
                    f.write(" PUBLIC '" + node.publicId + "' ")
            else:
                f.write(' SYSTEM ')

            if '"' not in node.systemId:
                f.write('"' + node.systemId + '"')
            else:
                f.write("'" + node.systemId + "'")

        if node.internalSubset:
            f.write(" [%s]" % node.internalSubset)
        elif node.entities.length or node.notations.length:
            f.write(' [\n')
            for i in range(node.entities.length):
                self.render(f, node.entities.item(i))
            for i in range(node.notations.length):
                self.render(f, node.notations.item(i))
            f.write(']')
        f.write('>\n')
       
    def renderNotation(self, f, node):
        st = "<!NOTATION %s" % node.nodeName
        if node.publicId:
            st = st + ' PUBLIC "%s"' % node.publicId
            if node.systemId:
                st = '%s "%s"' % (st, node.systemId)
        elif node.systemId:
            st = st + ' SYSTEM "%s"' % node.systemId
        f.write(st + '>\n')

    def renderDocumentFragment(self, f, node):
        for child in node.childNodes:
            self.render(f, child)
            
# regexps used by _translateCdata(),
# made global to compile once.
# see http://www.xml.com/axml/target.html#dt-character
ILLEGAL_LOW_CHARS = '[\x01-\x08\x0B-\x0C\x0E-\x1F]'
ILLEGAL_HIGH_CHARS = '\xEF\xBF[\xBE\xBF]'
# Note: Prolly fuzzy on this, but it looks as if characters from the
# surrogate block are allowed if in scalar form, which is encoded in UTF8 the
# same was as in surrogate block form
XML_ILLEGAL_CHAR_PATTERN = re.compile(
    '%s|%s' % (ILLEGAL_LOW_CHARS, ILLEGAL_HIGH_CHARS))
# the characters that we will want to turn into entrefs
# We must do so for &, <,  and > following ]].
# The xml parser has more leeway, but we're not the parser.
# http://www.xml.com/axml/target.html#dt-chardata
# characters that we must *always* turn to entrefs:
g_cdataCharPatternReq = re.compile('[&<]|]]>')
g_charToEntityReq = {
    '&': '&amp;',
    '<': '&lt;',
    ']]>': ']]&gt;',
    }
# characters that we must turn to entrefs in attr values:
g_cdataCharPattern = re.compile('[&<>"\']|]]>')
g_charToEntity = {
    '&': '&amp;',
    '<': '&lt;',
    '>': '&gt;',
    '"': '&quot;',
    "'": '&apos;',
    ']]>': ']]&gt;',
    }

def _translateCdata(characters, allEntRefs = None):
    """Translate characters into a legal format."""
    if not characters:
        return ''
    if allEntRefs: # translate all chars to entrefs; for attr value
        if g_cdataCharPattern.search(characters):
            new_string = g_cdataCharPattern.subn(
                lambda m, d=g_charToEntity: d[m.group()],
                characters)[0]
        else:
            new_string = characters
    else: # translate only required chars to entrefs
        if g_cdataCharPatternReq.search(characters):
            new_string = g_cdataCharPatternReq.subn(
                lambda m, d=g_charToEntityReq: d[m.group()],
                characters)[0]
        else:
            new_string = characters
    if XML_ILLEGAL_CHAR_PATTERN.search(new_string):
        new_string = XML_ILLEGAL_CHAR_PATTERN.subn(
            lambda m: '&#%i;' % ord(m.group()),
            new_string)[0]
    return new_string
    
def _translateCdataAttr(characters):
    """
    Translate attribute value characters into a legal format;
    return the value and the delimiter used.
    """
    if not characters:
        return '', '"'
    if '"' not in characters or "'" in characters:
        delimiter = '"'
        new_chars = _translateCdata(characters, allEntRefs = 1)
        new_chars = re.sub("&apos;", "'", new_chars)
    else:
        delimiter = "'"
        new_chars = _translateCdata(characters, allEntRefs = 1)
        new_chars = re.sub("&quot;", '"', new_chars)
    return new_chars, delimiter

def indentBlock(text, indent, line_length):
    words = text.split()
    lines = []
    i = 0
    while i < len(words):
        line = []
        while i < len(words) and indent + len(" ".join(line)) < line_length:
            line.append(words[i])
            i += 1
        if len(line) > 1 and indent + len(" ".join(line)) >= line_length:
            i -= 1
            line.pop()
        lines.append(" " * indent + " ".join(line))
    return '\n'.join(lines)

    



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import DOM
import xml.dom

from ComputedAttribute import ComputedAttribute

import string

_DOM_PROXY_FEATURES = ()

class DOMImplementationProxy:
    def __init__(self):
        self._domimplementation = DOM.theDOMImplementation

    def hasFeature(self, feature, version):
        feature = string.lower(feature)
        if (feature, version) in _DOM_PROXY_FEATURES:
            return 1
        return self._domimplementation.hasFeature(feature, version)

    def _createDOMDocumentType(self, qualifiedName, publicId, systemId):
        return self._domimplementation.createDocumentType(
            qualifiedName, publicId, systemId)

    def _createDOMDocument(self, namespaceURI, qualifiedName, docType=None):
        return self._domimplementation.createDocument(
            namespaceURI, qualifiedName, docType)

class DOMProxy:

    def __init__(self, node, persistentDoc=None):
        self._node = node
        self._persistentDoc = persistentDoc
        
    def getDOMObj(self):
        """Return the node without wrappers.
        """
        return self._node
    
    def getPersistentDoc(self):
        """Return a reference to a security friendly persistent document
        if we can, or None.  This can be None when OwnerDocument, if it
        exsists, is not None.  If you don't need to reach the document
        through the security checks, you don't need to use this method.
        """    
        # We do this because the userfolder container needs to be in the aq
        # context of the unwrapped returned object.
        if self._persistentDoc and getattr(
            self._persistentDoc, "_container", None):
            
            try:
                # If we can get to self, we can get to the persistent doc -
                # but restrictedTraverse is necessary for some reason for
                # what we return to be used in an acquisition chain
                return self._persistentDoc._container.restrictedTraverse(
                    self._persistentDoc.getPhysicalPath())
            except:
                pass
        return None

    def __setattr__(self, name, value):
        """Proxy DOM attribute writes, else write to our attribute.
        """
        if name in self._DOMAttrs:
            setattr(self._node, name, value)
        else:
            self.__dict__[name] = value
            # flag that we're dirty if we're persistent
            # FIXME: perhaps there's a better way to make those
            # test cases work?
            if hasattr(self, '_p_changed'):
                self._p_changed = 1
            
    def __nonzero__(self):
        "is this node true?"
        # FIXME: not sure this makes sense
        try:
            return self._node.__nonzero__()
        except:
            if self._node:
                return 1
            else:
                return 0
        
    # no need to avoid aq bugs fixed in Zope 2.3.1 so no __len__

class NodeProxy(DOMProxy):
    _DOMAttrs = ("nodeName", "attributes", "childNodes", "firstChild",
                 "lastChild", "localName", "namespaceURI", "nextSibling",
                 "previousSibling",
                 "nodeType", "nodeValue", "ownerDocument", "parentNode",
                 "prefix")
    
    ELEMENT_NODE = 1
    ATTRIBUTE_NODE = 2
    TEXT_NODE = 3
    CDATA_SECTION_NODE = 4
    ENTITY_REFERENCE_NODE = 5
    ENTITY_NODE = 6
    PROCESSING_INSTRUCTION_NODE = 7
    COMMENT_NODE = 8
    DOCUMENT_NODE = 9
    DOCUMENT_TYPE_NODE = 10
    DOCUMENT_FRAGMENT_NODE = 11
    NOTATION_NODE = 12
    
    def _get_nodeName(self):
        return self._node._get_nodeName()
    nodeName = ComputedAttribute(_get_nodeName)
    
    def _get_attributes(self):
        return self.wrapNamedNodeMap(self._node._get_attributes())
    attributes = ComputedAttribute(_get_attributes)

    def _get_childNodes(self):
        return self.wrapNodeList(self._node._get_childNodes())
    childNodes = ComputedAttribute(_get_childNodes)
    
    def _get_firstChild(self):
        return self.wrapDOMObj(self._node._get_firstChild())
    firstChild = ComputedAttribute(_get_firstChild)
    
    def _get_lastChild(self):
        return self.wrapDOMObj(self._node._get_lastChild())
    lastChild = ComputedAttribute(_get_lastChild)
    
    def _get_localName(self):
        return self._node._get_localName()
    localName = ComputedAttribute(_get_localName)
    
    def _get_namespaceURI(self):
        return self._node._get_namespaceURI()
    namespaceURI = ComputedAttribute(_get_namespaceURI)
    
    def _get_nextSibling(self):
        return self.wrapDOMObj(self._node._get_nextSibling())
    nextSibling = ComputedAttribute(_get_nextSibling)
    
    def _get_previousSibling(self):
        return self.wrapDOMObj(self._node._get_previousSibling())
    previousSibling = ComputedAttribute(_get_previousSibling)
    
    def _get_nodeType(self):
        return self._node._get_nodeType()
    nodeType = ComputedAttribute(_get_nodeType)
    
    def _get_nodeValue(self):
        return self._node._get_nodeValue()
    nodeValue = ComputedAttribute(_get_nodeValue)
        
    def _get_parentNode(self):
        return self.wrapDOMObj(self._node._get_parentNode())
    parentNode = ComputedAttribute(_get_parentNode)
    
    def _get_prefix(self):
        return self._node._get_prefix()
    prefix = ComputedAttribute(_get_prefix)
    
    def hasAttributes(self):
        return self._node.hasAttributes()

    def hasChildNodes(self):
        return self._node.hasChildNodes()

    def appendChild(self, newChild):
        self._node.appendChild(newChild._node)
        return newChild
    
    def insertBefore(self, newChild, refChild):
        self._node.insertBefore(newChild._node,
                                getattr(refChild, "_node", None))
        return newChild

    def removeChild(self, oldChild):
        self._node.removeChild(oldChild._node)
        return oldChild

    def replaceChild(self, newChild, oldChild):
        self._node.replaceChild(newChild._node, oldChild._node)
        return oldChild

    def normalize(self):
        self._node.normalize()

    def isSupported(self, feature, version):
        return self._node.isSupported(feature, version)

    def isSameNode(self, other):
        return self._node.isSameNode(other._node)

    def cloneNode(self, deep):
        return self.wrapDOMObj(self._node.cloneNode(deep))

    def __cmp__(self, other):
        try:
            return cmp(self._node, other._node)
        except AttributeError:
            return 1
    
    def _get_ownerDocument(self):
        """
        Return the DOM document - the persistent document that was instantiated
        if we can, otherwise another generated proxy node.
        """
        return (self.getPersistentDoc() or
                self.wrapDOMObj(self._node._get_ownerDocument()))

    ownerDocument = ComputedAttribute(_get_ownerDocument)

    def __hash__(self):
        return self._node.__hash__()
    
class NodeListProxy(DOMProxy):
    _DOMAttrs = ("length",)

    def __len__(self):
        return self._node.__len__()

    def _get_length(self):
        return self._node._get_length()

    length = ComputedAttribute(_get_length)

    def __nonzero__(self):
        return self._node.__nonzero__()

    def __setitem__(self, i, newChild):
        self._node.__setitem__(i, newChild._node)

    def __delitem__(self, i):
        self._node.__delitem__(i)

    def __getitem__(self, i):
        return self.wrapDOMObj(self._node.__getitem__(i))

    def __getslice__(self, i, j):
        return self.wrapNodeList(self._node.__getslice__(i, j))

    def item(self, i):
        return self.wrapDOMObj(self._node.item(i))

    def count(self, value):
        return self._node.count(value._node)
    
    def index(self, value):
        return self._node.index(value._node)
    
class NamedNodeMapProxy(DOMProxy):
    _DOMAttrs = ("length", "keys")

    def getDOMObj(self):
        """Return the Node without any of our wrappers
        """
        return self._node

    def __len__(self):
        return self._node.__len__()
    
    def _get_length(self):
        return self._node._get_length()
    length = ComputedAttribute(_get_length)

    def setNamedItem(self, arg):
        return self.wrapDOMObj(self._node.setNamedItem(arg._node))

    def setNamedItemNS(self, arg):
        return self.wrapDOMObj(self._node.setNamedItemNS(arg._node))

    def removeNamedItem(self, name):
        return self.wrapDOMObj(self._node.removeNamedItem(name))

    def removeNamedItemNS(self, namespaceURI, localName):
        return self.wrapDOMObj(self._node.removeNamedItemNS(namespaceURI,
                                                            localName))

    def get(self, name, default=None):
        node = self._node.getNamedItem(name)
        if node is None:
            return default
        else:
            return self.wrapDOMObj(node)

    def has_key(self, name):
        return self._node.has_key(name)

    def item(self, i):
        return self.wrapDOMObj(self._node.item(i))

    def getNamedItem(self, name):
        return self.wrapDOMObj(self._node.getNamedItem(name))

    def getNamedItemNS(self, namespaceURI, localName):
        return self.wrapDOMObj(self._node.getNamedItemNS(namespaceURI,
                                                         localName))

    def __setitem__(self, name, node):
        self._node.__setitem__(name, node._node)

    def __getitem__(self, name):
        return self.wrapDOMObj(self._node.__getitem__(name))

    def __delitem__(self, name):
        self._node.__delitem__(name)

    def keys(self):
        return self._node.keys()
    
    def values(self):
        return map(self.wrapDOMObj, self._node.values())

    def items(self):
        result = []
        for key, value in self.getDOMObj().items():
            result.append((key, self.wrapDOMObj(value)))
        return result

class DocumentFragmentProxy(NodeProxy):
    pass
    
class ElementProxy(NodeProxy):
    _DOMAttrs = NodeProxy._DOMAttrs + ("tagName", 'element_id')

    def _get_tagName(self):
        return self._node._get_tagName()
    
    tagName = ComputedAttribute(_get_tagName)

    def _get_elementId(self):
        return self._node._get_elementId()

    elementId = ComputedAttribute(_get_elementId)
    
    def getAttribute(self, name):
        return self._node.getAttribute(name)

    def getAttributeNS(self, namespaceURI, localName):
        return self._node.getAttributeNS(namespaceURI, localName)

    def getAttributeNode(self, name):
        return self.wrapDOMObj(self._node.getAttributeNode(name))

    def getAttributeNodeNS(self, namespaceURI, localName):
        return self.wrapDOMObj(self._node.getAttributeNodeNS(namespaceURI,
                                                           localName))

    def getElementsByTagName(self, name):
        return self.wrapNodeList(self._node.getElementsByTagName(name))
    
    def getElementsByTagNameNS(self, namespaceURI, localName):
        return self.wrapNodeList(self._node.getElementsByTagNameNS(namespaceURI,
                                                                   localName))

    def hasAttribute(self, name):
        return self._node.hasAttribute(name)

    def hasAttributeNS(self, namespaceURI, localName):
        return self._node.hasAttributeNS(namespaceURI, localName)

    def removeAttribute(self, name):
        self._node.removeAttribute(name)

    def removeAttributeNS(self, namespaceURI, localName):
        self._node.removeAttributeNS(namespaceURI, localName)

    def removeAttributeNode(self, oldAttr):
        return self.wrapDOMObj(self._node.removeAttributeNode(oldAttr._node))

    def setAttribute(self, name, value):
        self._node.setAttribute(name, value)

    def setAttributeNS(self, namespaceURI, qualifiedName, value):
        self._node.setAttributeNS(namespaceURI, qualifiedName, value)

    def setAttributeNode(self, newAttr):
        return self.wrapDOMObj(self._node.setAttributeNode(newAttr._node))

    def setAttributeNodeNS(self, newAttr):
        return self.wrapDOMObj(self._node.setAttributeNodeNS(newAttr._node))

class CharacterDataProxy(NodeProxy):
    _DOMAttrs = NodeProxy._DOMAttrs + ("data", "length")
    
    def __len__(self):
        return self._node.__len__()

    def _get_data(self):
        return self._node._get_data()

    data = ComputedAttribute(_get_data)

    def _set_data(self, value):
        self._node._set_data(value)
        
    def _get_length(self):
        return self._node._get_length()
    
    length = ComputedAttribute(_get_length)

    def appendData(self, arg):
        self._node.appendData(arg)

    def deleteData(self, offset, count):
        self._node.deleteData(offset, count)

    def insertData(self, offset, arg):
        self._node.insertData(offset, arg)

    def replaceData(self, offset, count, arg):
        self._node.replaceData(offset, count, arg)

    def substringData(self, offset, count):
        return self._node.substringData(offset, count)
        
class TextProxy(CharacterDataProxy):
    def splitText(self, offset):
        return self.wrapDOMObj(self._node.splitText(offset))

class CDATASectionProxy(TextProxy):
    pass

class CommentProxy(CharacterDataProxy):
    pass

class ProcessingInstructionProxy(NodeProxy):
    _DOMAttrs = NodeProxy._DOMAttrs + ("target", "data")

    def _get_target(self):
        return self._node._get_target()

    target = ComputedAttribute(_get_target)

    def _get_data(self):
        return self._node._get_data()

    data = ComputedAttribute(_get_data)

    def _set_data(self, value):
        self._node._set_data(value)
        
class AttrProxy(NodeProxy):
    _DOMAttrs = NodeProxy._DOMAttrs + ("name", "value", "specified",
                                       "ownerElement")

    def _get_name(self):
        return self._node._get_name()

    name = ComputedAttribute(_get_name)
    
    def _get_value(self):
        return self._node._get_value()
    # setting should also be okay
    value = ComputedAttribute(_get_value)

    def _set_value(self, value):
        self._node._set_value(value)
        
    def _get_specified(self):
        return self._node._get_specified()

    specified = ComputedAttribute(_get_specified)
    
    def _get_ownerElement(self):
        return self.wrapDOMObj(self._node._get_ownerElement())

    ownerElement = ComputedAttribute(_get_ownerElement)
    
class DocumentProxy(NodeProxy):
    _DOMAttrs = NodeProxy._DOMAttrs + ("doctype", "documentElement")
    
    def _get_doctype(self):
        return self.wrapDOMObj(self._node._get_doctype())

    doctype = ComputedAttribute(_get_doctype)

    def _get_documentElement(self):
        return self.wrapDOMObj(self._node._get_documentElement())

    documentElement = ComputedAttribute(_get_documentElement)

    # _get_implementation defined in subclass
    
    def createAttribute(self, name):
        return self.wrapDOMObj(self._node.createAttribute(name))

    def createAttributeNS(self, namespaceURI, qualifiedName):
        return self.wrapDOMObj(self._node.createAttributeNS(namespaceURI,
                                                            qualifiedName))

    def createCDATASection(self, data):
        return self.wrapDOMObj(self._node.createCDATASection(data))

    def createComment(self, data):
        return self.wrapDOMObj(self._node.createComment(data))

    def createDocumentFragment(self):
        return self.wrapDOMObj(self._node.createDocumentFragment())

    def createElement(self, tagName):
        return self.wrapDOMObj(self._node.createElement(tagName))

    def createElementNS(self, namespaceURI, qualifiedName):
        return self.wrapDOMObj(self._node.createElementNS(namespaceURI,
                                                           qualifiedName))
    def createEntityReference(self, name):
        return self.wrapDOMObj(self._node.createEntityReference(name))

    def createProcessingInstruction(self, target, data):
        return self.wrapDOMObj(self._node.createProcessingInstruction(target,
                                                                      data))
    def createTextNode(self, data):
        return self.wrapDOMObj(self._node.createTextNode(data))

    def getElementById(self, elementId):
        return self.wrapDOMObj(self._node.getElementById(elementId))

    def getElementsByTagName(self, tagName):
        return self.wrapNodeList(self._node.getElementsByTagName(tagName))

    def getElementsByTagNameNS(self, namespaceURI, localName):
        return self.wrapNodeList(self._node.getElementsByTagNameNS(namespaceURI,
                                                                   localName))
    
    def importNode(self, importedNode, deep):
        return self.wrapDOMObj(self._node.importNode(importedNode._node,
                                                     deep))

    def createNodeIterator(self, root, whatToShow, filter,
                           entityReferenceExpansion):
        # FIXME: needs wrapper!
        return self._node.createNodeIterator(root, whatToShow, filter,
                                             entityReferenceExpansion)

    def createTreeWalker(self, root, whatToShow, filter,
                         entityReferenceExpansion):
        # FIXME: needs wrapper!
        return self._node.createTreeWalker(root, whatToShow, filter,
                                           entityReferenceExpansion)
    
class EntityProxy(NodeProxy):
    _DOMAttrs = NodeProxy._DOMAttrs + ("publicId", "systemId", "notationName")

    def _get_publicId(self):
        return self._node._get_publicId()

    publicId = ComputedAttribute(_get_publicId)

    def _get_systemId(self):
        return self._node._get_systemId()

    systemId = ComputedAttribute(_get_systemId)

    def _get_notationName(self):
        return self._node._get_notationName()

    notationName = ComputedAttribute(_get_notationName)

class EntityReferenceProxy(NodeProxy):
    pass

class NotationProxy(NodeProxy):
    _DOMAttrs = NodeProxy._DOMAttrs + ("publicId", "systemId")

    def _get_publicId(self):
        return self._node._get_publicId()

    publicId = ComputedAttribute(_get_publicId)

    def _get_systemId(self):
        return self._node._get_systemId()

    systemId = ComputedAttribute(_get_systemId)
    
class DocumentTypeProxy(NodeProxy):
    _DOMAttrs = NodeProxy._DOMAttrs + ("publicId", "systemId",
                                       "name", "entities", "notations",
                                       "internalSubset")
    def _get_entities(self):
        return self.wrapNamedNodeMap(self._node._get_entities())

    entities = ComputedAttribute(_get_entities)
    
    def _get_internalSubset(self):
        return self._node._get_internalSubset()

    internalSubset = ComputedAttribute(_get_internalSubset)
    
    def _get_name(self):
        return self._node._get_name()

    name = ComputedAttribute(_get_name)
    
    def _get_notations(self):
        return self.wrapNamedNodeMap(self._node._get_notations())

    notations = ComputedAttribute(_get_notations)
    
    def _get_publicId(self):
        return self._node._get_publicId()

    publicId = ComputedAttribute(_get_publicId)
    
    def _get_systemId(self):
        return self._node._get_systemId()

    systemId = ComputedAttribute(_get_systemId)
    
